@@ -268,7 +268,7 @@ static bool ads_keytab_verify_ticket(krb5_context context,
 		}
 	}
 
-	SAFE_FREE(entry_princ_s);
+	TALLOC_FREE(entry_princ_s);
 
 	{
 		krb5_keytab_entry zero_kt_entry;
