@@ -1365,7 +1365,7 @@ static NTSTATUS winbindd_dual_pam_auth_samlogon(struct winbindd_domain *domain,
 			domain->can_do_validation6 = false;
 		}
 
-		logon_fn = contact_domain->can_do_samlogon_ex
+		logon_fn = (contact_domain->can_do_samlogon_ex && domain->can_do_validation6)
 			? rpccli_netlogon_sam_network_logon_ex
 			: rpccli_netlogon_sam_network_logon;
 
@@ -1558,7 +1558,9 @@ enum winbindd_result winbindd_dual_pam_auth(struct winbindd_domain *domain,
 	parse_domain_user(mapped_user, name_domain, name_user);
 
 	if ( mapped_user != state->request->data.auth.user ) {
-		fstr_sprintf( domain_user, "%s\\%s", name_domain, name_user );
+		fstr_sprintf( domain_user, "%s%c%s", name_domain,
+			*lp_winbind_separator(),
+			name_user );
 		safe_strcpy( state->request->data.auth.user, domain_user,
 			     sizeof(state->request->data.auth.user)-1 );
 	}
@@ -1989,7 +1991,7 @@ enum winbindd_result winbindd_dual_pam_auth_crap(struct winbindd_domain *domain,
 			domain->can_do_validation6 = false;
 		}
 
-		logon_fn = contact_domain->can_do_samlogon_ex
+		logon_fn = (contact_domain->can_do_samlogon_ex && domain->can_do_validation6)
 			? rpccli_netlogon_sam_network_logon_ex
 			: rpccli_netlogon_sam_network_logon;
 
